BeginPackage["HEPMath`FeynIntegrals`",
  {"HEPMath`Tensors`", "HEPMath`Lorentz`", "HEPMath`PaVe`"}]

DenCancel::usage = "DenCancel[expr] cancels propagator denominators in expr 
against squared momenta in the numerator. All propagator denominators must be
written in the form Den[q, m]. The squared momenta in the numerator must match
the momentum in the denominator up to a minus sign. For example, DenCancel
can simplify Sqr[l+p] Den[l+p, m] and Sqr[-l] Den[l, m] but not
Sqr[l] Den[l+p, m] or SDot[l, p] Den[l+p, m]. DenCancel[expr, patt] leaves any
subexpressions free of patter patt untouched."

FeynmanIntegrate::usage = "FeynmanIntegrate[expr, {l1, l2, ...}, x, muR, del]
integrates expr over the loop momenta l1, l2, etc. and expresses the result as
an integral over Feynman parameters x[1], x[2], etc. Only scalar integrals can
be computed in this way, i.e. the loop momenta l1, l2, ... are only allowed to
appear in propagator denominators (Den expressions). muR is the renormalization
scale and del is the symbol used for the infinitesimal imaginary part of the
propagator denominators. Omitting it is equivalent to setting it to zero.
The result is returned as a FeynmanIntegral expression."

FeynmanIntegrate::indep = "Integrand is independent of loop momenta `1`."

FeynmanIntegrate::novec = "Loop momenta `1` must be Lorentz vectors."

FeynmanIntegral::usage =
"FeynmanIntegral[{G, M, U, F}, {x1, ..., xn}] represents the
Feynman integral

    G \\int dx_1...dx_n delta(1-x_1-...-x_n) M U F

where G is a pre-factor, M is a monomial in the x_i arising from denominators
raised to some power different form 1, and U and F are polynomials in the x_i
raised to some non-integer power depending on the number Dim of space-time
dimensions."

EliminateFeynmanParameter::usage =
"EliminateFeynmanParameter[f, xi, xlist] eliminates the Feynman
parameter xi in the Feynman integral f by integrating out the delta function. f
must be a FeynmanIntegral expression. The result is a HEPIntegral expression
representing the integral over the remaining Feynman parameters. The xlist
argument is optional. If given it must contain the remaining Feynman parameters
in some order. Integration limits are chosen in such a way that the limits of
parameter xj only depend on parameters appearing before xj in xlist. If the
second and third argument are omitted the function automatically chooses a
parameter to eliminate."

EliminateFeynmanParameter::fail = "Invalid call of EliminateFeynmanParameter
with argument list `1`."


Begin["`Private`"]

(* DenCancel *)

canonicalSquares[ex_, patt_] := Block[{res=ex},
  res = res /. Den[q_ /; !FreeQ[q, patt], m_] :> Den[Expand[q], m];
  res = res /. Sqr[q_ /; !FreeQ[q, patt]] :> Sqr[Expand[q]];
  res = res /. Den[q_Plus, m_] :> Den[Expand[-q], m] /;
    MatchQ[First[q], (n_?NumberQ /; n < 0) _] && !FreeQ[q, patt];
  res = res /. Den[q_Times, m_] :> Den[-q, m] /;
    MatchQ[First[q], n_?NumberQ /; n < 0] && !FreeQ[q, patt];
  res = res /. Sqr[q_Plus] :> Sqr[Expand[-q]] /;
    MatchQ[First[q], (n_?NumberQ /; n < 0) _] && !FreeQ[q, patt];
  res = res /. Sqr[q_Times] :> Sqr[-q] /;
    MatchQ[First[q], n_?NumberQ /; n < 0] && !FreeQ[q, patt];

  res
]

denCancel[ex_, patt_] := Expand[ex, _Den] /. 
  (Den[q_ /; !FreeQ[q, patt], m_]^(n_Integer) Sqr[q_]^(k_Integer) :>
    Sqr[q]^(k-1) Den[q, m]^(n-1) + m^2 Sqr[q]^(k-1) Den[q, m]^n /;
    k > 1 && n > 1) /.
  (Den[q_ /; !FreeQ[q, patt], m_] Sqr[q_]^(m_Integer) :>
    Sqr[q]^(m-1) + m^2 Sqr[q]^(m-1) Den[q, m] /; k > 1) /.
  (Den[q_ /; !FreeQ[q, patt], m_]^(n_Integer) Sqr[q_] :>
    Den[q, m]^(n-1) + m^2 Den[q, m]^n /; n > 1) /.
  Den[q_ /; !FreeQ[q, patt], m_] Sqr[q_] :> 1 + m^2 Den[q, m]
  
DenCancel[ex_, patt_] := 
  FixedPoint[denCancel[#, patt]&, canonicalSquares[ex, patt]]

DenCancel[ex_] := DenCancel[ex, _]


(* FeynmanIntegrate *)

minorMatrix[m_List?MatrixQ] := Map[Reverse, Minors[m], {0, 1}]

cofactorMatrix[m_List?MatrixQ] :=
  MapIndexed[#1 (-1)^(Plus @@ #2)&, minorMatrix[m],{2}]

feynmanParametrize[dd_, ll_, x_, muR_, dl_] := Block[
  {den, terms, coeff, mm, qq, jj, lpatt, pos, dum,
   ff, uu, nu, nui, nl, nx, mmcf},

  lpatt = Alternatives @@ ll;
  den = dd;
  nx = Length[den];
  nui = Which[Head[#] === Den,     1,
              MatchQ[#, (_Den)^_], #[[2]],
              True,                Message[FeynmanIntegrate::fail, dd]]& /@ den;
  nu = Plus @@ nui;
  nl = Length[ll];

  den = den /. (d_Den)^_ -> d;
  den = Sum[x[i] den[[i]], {i, nx}];
  den = den /. Den[l_, m_] :> -Sqr[l] + m^2;
  den = HEPExpand[den, lpatt|_x];
  terms = HEPCollectListed[den, lpatt, CombineTensorStructures->False];

  {mm, terms} = HEPGroupTerms[terms,
    MatchQ[#, (HoldPattern[SDot[dum, dum]|Sqr[dum]] /. dum -> lpatt)]&];
  {qq, jj} = HEPGroupTerms[terms,
    MatchQ[#, HoldPattern[SDot[dum, _]|SDot[_, dum]] /. dum -> lpatt]&];

  mm = Table[
    -HEPCoefficient[mm, SDot[ ll[[i]], ll[[j]] ]]/If[i===j, 1, 2],
    {i, Length[ll]}, {j, Length[ll]}];
  qq = Table[
    Plus @@ MapThread[(Replace[#1, SDot[l1, p_] -> p] #2/2)&,
                      HEPSelectTerms[qq, MatchQ[#, SDot[l1, _]]&]],
    {l1, ll}];
  jj = Dot @@ jj - I dl;

  uu = Det[mm];
  mmcf = cofactorMatrix[mm];
  ff = uu*jj + Sum[SDot[qq[[i]], qq[[j]]] mmcf[[i,j]], {i, nl}, {j, nl}];

  FeynmanIntegral[{
    (-1)^nu I^nl Pi^(nl Dim/2) (2 Pi muR)^(4 nl - nl Dim) *
      Gamma[nu - nl Dim/2]/(Times @@ (Gamma /@ nui)),
    Product[x[i]^(nui[[i]]-1), {i, Length[nui]}],
    uu^(nu - (nl+1) Dim/2), ff^(nl Dim/2 - nu)},
    Table[x[i], {i, nx}]]
]

unZip[l_List, f_] := Block[{true={},false={}},
  Scan[If[f[#]===True, AppendTo[true, #], AppendTo[false, #]]&, l];
  {true,false}
]

feynmanFactorize[{fac___, {{}, {}}}] := {fac}

feynmanFactorize[{fac___, {dd_, ll_}}] := Block[
  {den, denr, den1, loop, loopr, loop1},

  loop = {First[ll]};
  loopr = Rest[ll];
  {den, denr} = unZip[dd, !FreeQ[#, Alternatives@@loop]&];
  {loop1, loopr} = unZip[loopr, !FreeQ[den, #]&];
  While[loop1 =!= {},
    loop = Join[loop, loop1];
    {den1, denr} = unZip[denr, !FreeQ[#, Alternatives@@loop]&];
    den = Join[den, den1];
    {loop1, loopr} = unZip[loopr, !FreeQ[den, #]&];
  ];
  
  feynmanFactorize[{fac, {den, loop}, {denr, loopr}}]
]

feynmanIntegrate[dd_, ll_, x_, muR_, dl_] := Block[{den},
  If[Or @@ ((HEPTensorSignature[#] =!= {Lorentz})& /@ ll),
    Message[FeynmanIntegrate::novec, ll]; Abort[]];
  If[FreeQ[dd, Alternatives@@ll],
    Message[FeynmanIntegrate::indep, ll]; ComplexInfinity
    , (* else *)
    den = If[Head[dd] === Times, List @@ dd, {dd}];
    den = feynmanFactorize[{{den, ll}}];
    Times @@ (feynmanParametrize[#[[1]], #[[2]], x, muR, dl]& /@ den)
  ]
]

FeynmanIntegrate[ex_, ll_List, x_, muR_, dl_] := Block[{den, coeff, lpatt},
  lpatt = Alternatives @@ ll;
  {den, coeff} = HEPCollectListed[ex, Den[l_ /; !FreeQ[l, lpatt], _],
                                  CombineTensorStructures -> False];
  den = feynmanIntegrate[#, ll, x, muR, dl]& /@ den;
  den.coeff
]

FeynmanIntegrate[ex_, l_, x_, muR_, dl_] := FeynmanIntegrate[ex, {l}, x, muR, dl]

FeynmanIntegrate[ex_, l_, x_, muR_] := FeynmanIntegrate[ex, l, x, muR, 0]


(* EliminateFeynmanParameter *)

EliminateFeynmanParameter[
    FeynmanIntegral[{gg_, mm_, uu_, ff_}, xx_List], xi_, xlist_List] := Block[
  {xmax, xvar, integ},

  If[MemberQ[xlist, xi] || Sort[Append[xlist, xi]] =!= Sort[xx],
    Message[EliminateFeynmanParameter::fail,
      {FeynmanIntegral[{gg, mm, uu, ff}, xx], xi, xlist}]; Abort[]];

  integ = mm uu ff /. xi -> Expand[1 - (Plus @@ xlist)];
  var = Table[{xlist[[i]], 0, Expand[1 - (Plus @@ Take[xlist, i-1])]},
              {i, Length[xlist]}];
  gg HEPIntegral[integ, Sequence@@var]
]

EliminateFeynmanParameter[FeynmanIntegral[f_, xx_List], xi_] :=
  EliminateFeynmanParameter[FeynmanIntegral[f, xx], xi, DeleteCases[xx, xi]]

EliminateFeynmanParameter[FeynmanIntegral[f_, {xx___, x_}]] :=
  EliminateFeynmanParameter[FeynmanIntegral[f, {xx, x}], x, {xx}]


End[]

EndPackage[]
